"""A module parsing arguments and putting them together."""
from typing import Any, Dict, Iterable, List, Type, Union, cast, get_args, get_origin

from eppic.input.input_processor import ParsedCommandArgument

from .argument_registry import ArgumentRegistry
from .arguments.argument import Argument, ArgumentField, ArgumentLabelSuggestion


def _unpack_required_type(argument_type: Type) -> Type:
    """Turn a potentially optional type into a type.

    Args:
        argument_type: the type to unpack.

    Returns: the unpacked type

    """
    args = list(get_args(argument_type))
    origin = get_origin(argument_type)

    if origin == Union and type(None) in args:
        args.remove(type(None))
        argument_type = args.pop(0)

        for more in args:
            argument_type = cast(Type, Union[argument_type, more])

        return argument_type

    return argument_type


class ArgumentParser:
    """A class responsible for parsing arguments."""

    def __init__(self, fields: Iterable[ArgumentField], registry: ArgumentRegistry) -> None:
        """Save arguments and initialize arguments list.

        Args:
            fields: an iterable of argument fields.
            registry: an ArgumentRegistry instance containing all the registered arguments.

        """
        self._fields = tuple(fields)
        self._registry = registry
        self._init_arguments_list()

    def _create_argument(self, field: ArgumentField) -> Argument:
        """Get an argument from registry and initialize it.

        Args:
            field: Argument field.

        Returns: an Argument instance.

        """
        argument_type = _unpack_required_type(field.type)
        argument_class = self._registry.get_argument(argument_type)

        return argument_class(ArgumentField(field.name, argument_type, bool(field.required)), self._create_argument)

    def _init_arguments_list(self) -> None:
        """Initialize the necessary properties before every parsing."""
        self._arguments_list = [self._create_argument(f) for f in self._fields]

    def parse_arguments(self, arguments: Iterable[ParsedCommandArgument]) -> None:
        """Parse arguments from user input (strings) to the requrested types."""
        self._init_arguments_list()
        for label, value in arguments:
            if label is None:
                label = self.get_next_argument_label()

            if value is None:
                raise ValueError(f"Argument {label} didn't recieve a value.")

            self._parse_argument(label, value)

    def _parse_argument(self, label: List[str], value: str) -> None:
        """Parse a single argument.

        Args:
            label: the label of the argument.
            value: the value of the arguments.

        """
        for argument in self._arguments_list:
            if argument.recognise_label(label):
                argument.add_value(label, value)
                return

        raise ValueError(f"Argument {label} not found")

    def get_next_argument_label(self) -> List[str]:
        """Get next available argument's label."""
        available = self.get_available_labels()
        if not available:
            raise ValueError("Too many arguments recieved.")

        return available[0][0]

    def get_available_labels(self) -> List[ArgumentLabelSuggestion]:
        """Get available argument labels, potentially with values, and whether they're required."""
        ret = []
        for argument in self._arguments_list:
            ret += argument.available_argument_labels()

        return ret

    def get_parsed_values(self) -> Dict[str, Any]:
        """Get previously parsed values from the arguments."""
        ret = {}
        for argument in self._arguments_list:
            for name, value in argument.get_value().items():
                ret[name] = value

        return ret
