"""Generic Argument class."""
from __future__ import annotations

from typing import Any, Callable, Dict, List, NamedTuple, Optional, Type


class ArgumentField(NamedTuple):
    """Representation of argument data, similar to `dataclass.Field`.

    Attributes:
        name: Name of the argument.
        type: Type of the argument. May contain `Union`.
        required: Whether the argument is required.
    """

    name: str
    type: Type
    required: bool


class ArgumentLabelSuggestion(NamedTuple):
    """A representation of an argument label suggestion."""

    labels: List[str]
    values: Optional[List[str]]
    required: bool


class Argument:
    """A generic representation of an argument.

    Works with all types, such that `the_type(x)`, where `x` is a string,
    returns the correct value.

    """

    def __init__(
            self,
            field: ArgumentField,
            factory: Callable[[ArgumentField], Argument]
    ):
        """Save the basic info about the argument.

        Args:
            field: Argument field.
            factory: the factory function creating potential subarguments.

        """
        self._arg_label = field.name
        self._arg_type = field.type
        self._required = field.required
        self._factory = factory

        self._value: Optional[Any] = None
        self._invalid = False

    def get_subarguments_full_label(self, sub_label: List[str]) -> List[str]:
        """Return the sub_label with self._arg_label added to the end."""
        return [self._arg_label] + sub_label

    def get_value(self) -> Dict[str, Any]:
        """Get value of the argument, with the correct type."""
        if self._invalid:
            raise TypeError(f"An invalid value was supplied to the argument {self._arg_label}.")

        if self._value is None and self._required:
            raise ValueError(f"No value for {self._arg_label}.")

        return {self._arg_label: self._value}

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        if self.has_value:
            raise ValueError(f"Value for {self._arg_label} is already provided!")
        try:
            self._value = self._arg_type(value)
        except ValueError:
            self._invalid = True
            raise TypeError(f"Supplied value to argument {self._arg_label} is of incorrect type.")

    def recognise_label(self, labels: List[str]) -> bool:
        """Find out whether the labels are yours or one of your subarguments'.

        Args:
            labels: the labels of the lowest-level subargument, split by levels.
                            (highest-argument level first)

        Returns:
            True if `labels` are my (subargument's) labels.

        """
        return len(labels) == 1 and labels[0] == self._arg_label

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        if not self.has_value:
            return [ArgumentLabelSuggestion([self._arg_label], None, self._required)]
        return []

    @property
    def has_value(self) -> bool:
        """Whether or not the argument already has value.

        Useful for some more complicated argument types.

        """
        return self._value is not None
