"""Handle printing of output."""
import datetime
import json
import sys
from dataclasses import asdict, is_dataclass
from enum import Enum
from pprint import PrettyPrinter
from typing import Any, List, Sequence

import yaml
from epplib.responses.base import Result
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import PygmentsTokens
from pygments import lex
from pygments.lexers import YamlLexer

INDENT_MULTIPLIER = 4


class EppicSafeDumper(yaml.SafeDumper):
    """Custom YAML safe dumper with eppic specifics."""

    def represent_enum(self, member: Enum) -> Any:
        """Represent enum members as their values."""
        return self.represent_data(member.value)


EppicSafeDumper.add_multi_representer(str, EppicSafeDumper.represent_str)
EppicSafeDumper.add_multi_representer(Enum, EppicSafeDumper.represent_enum)


class OutputHandler:
    """Prints output."""

    def __init__(self, syntax_highlight: bool = True) -> None:
        """Create a PP instance."""
        self.pp = PrettyPrinter(indent=INDENT_MULTIPLIER)
        self._syntax_highlight = syntax_highlight

    def _serialize(self, data_list: Sequence[Any]) -> List[Any]:
        """Serialize res_data objects to simple types."""
        result = []
        for data in data_list:
            if is_dataclass(data):
                result.append(asdict(data))
            else:
                result.append(data)
        return result

    def _output(self, msg: str, code: int, data_list: List[Any]) -> None:
        print(f"{msg} ({code})\n")
        data = yaml.dump_all(data_list, Dumper=EppicSafeDumper, allow_unicode=True)
        if self._syntax_highlight:
            tokens = list(lex(data, lexer=YamlLexer()))
            print_formatted_text(PygmentsTokens(tokens))

        else:
            print(data)

    def output(self, res: Result) -> None:
        """Output the Result in res."""
        res_data = res.res_data if res.res_data is not None else []
        self._output(res.msg, res.code, self._serialize(res_data))


class JsonOutputHandler(OutputHandler):
    """Outputs JSON. Multiple responses are split by newlines."""

    def _output(self, msg: str, code: int, data_list: List[Any]) -> None:
        res = {"msg": msg, "code": code, "data": data_list}
        json.dump(res, sys.stdout, sort_keys=False, default=self._json_converter, ensure_ascii=False)
        print("")

    def _json_converter(self, obj: Any) -> Any:  # pragma: no branch
        """Convert complex types to JSON format."""
        if isinstance(obj, (datetime.datetime, datetime.date)):
            obj = obj.isoformat()

        # only happens when we encounter an unknown object.
        return obj  # pragma: no cover


class YamlOutputHandler(OutputHandler):
    """Outputs YAML. Multiple responses are split into multiple documents."""

    def _output(self, msg: str, code: int, data_list: Sequence[Any]) -> None:
        res = {"msg": msg, "code": code, "data": data_list}
        yaml.dump(res, sys.stdout, sort_keys=False, explicit_start=True, explicit_end=True, allow_unicode=True,
                  Dumper=EppicSafeDumper)
