import datetime
import io
import json
from dataclasses import dataclass
from enum import Enum
from typing import List, cast
from unittest import TestCase
from unittest.mock import Mock, patch

import yaml
from epplib.responses.base import Result
from testfixtures import OutputCapture

from eppic.output import JsonOutputHandler, OutputHandler, YamlOutputHandler


class TestEnum(int, Enum):
    TEST = 42


@dataclass
class EnumResData:
    member: TestEnum = TestEnum.TEST


@dataclass
class DummyResData:
    foo: str
    bar: str


@dataclass
class DummyResponse:
    code: int
    msg: str
    res_data: List[DummyResData]


class TestOutputHandler(TestCase):

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_output_empty(self, mock_out):
        res = cast(Result, DummyResponse(1000, "msg", []))
        oh = OutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            "msg (1000)\n\n"
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    @patch('eppic.output.print_formatted_text')
    def test_output_syntax_on(self, mock_format_print, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = OutputHandler()
        oh.output(res)
        mock_format_print.assert_called_once()

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_output_multiple_res_data(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = OutputHandler(False)
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(), """msg (1000)

bar: asdf
foo: asdf
---
bar: bar
foo: foo

"""
        )

    def test_output_enum(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=[EnumResData()])
        handler = OutputHandler(False)

        with OutputCapture() as output:
            handler.output(result)

        output.compare("Quagaars! (1000)\n\nmember: 42")

    def test_output_not_dataclass(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Fried chicken'])
        handler = OutputHandler(False)

        with OutputCapture() as output:
            handler.output(result)

        output.compare("Quagaars! (1000)\n\nFried chicken\n...")

    def test_output_unicode(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Pečené kuře'])
        handler = OutputHandler(False)

        with OutputCapture() as output:
            handler.output(result)

        output.compare("Quagaars! (1000)\n\nPečené kuře\n...")

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_yaml(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = YamlOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            """---
msg: msg
code: 1000
data:
- foo: asdf
  bar: asdf
- foo: foo
  bar: bar
...
"""
        )

    def test_yaml_enum(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=[EnumResData()])
        handler = YamlOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        result_data = {"msg": "Quagaars!", "code": 1000, "data": [{"member": 42}]}
        self.assertEqual(yaml.safe_load(output.captured), result_data)

    def test_yaml_not_dataclass(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Fried chicken'])
        handler = YamlOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        result_data = {"msg": "Quagaars!", "code": 1000, "data": ['Fried chicken']}
        self.assertEqual(yaml.safe_load(output.captured), result_data)

    def test_yaml_unicode(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Pečené kuře'])
        handler = YamlOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        self.assertIn('Pečené kuře', output.captured)
        result_data = {"msg": "Quagaars!", "code": 1000, "data": ['Pečené kuře']}
        self.assertEqual(yaml.safe_load(output.captured), result_data)

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "asdf", "bar": "asdf"}, {"foo": "foo", "bar": "bar"}]}\n'
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json_with_newline(self, mock_out):
        res_data = [
            DummyResData("asdf", "a\nsdf"),
            DummyResData("foo", "bar")
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "asdf", "bar": "a\\nsdf"}, {"foo": "foo", "bar": "bar"}]}\n'
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json_with_datetime(self, mock_out):
        res_data = [
            DummyResData("foo", cast(str, datetime.date(day=10, month=10, year=2020)))
        ]
        res = cast(Result, DummyResponse(1000, "msg", res_data))
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "foo", "bar": "2020-10-10"}]}\n'
        )

    def test_json_enum(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=[EnumResData()])
        handler = JsonOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        result_data = {"msg": "Quagaars!", "code": 1000, "data": [{"member": 42}]}
        self.assertEqual(json.loads(output.captured), result_data)

    def test_json_not_dataclass(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Fried chicken'])
        handler = JsonOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        result_data = {"msg": "Quagaars!", "code": 1000, "data": ['Fried chicken']}
        self.assertEqual(json.loads(output.captured), result_data)

    def test_json_unicode(self):
        result = Mock(code=1000, msg="Quagaars!", res_data=['Pečené kuře'])
        handler = JsonOutputHandler()

        with OutputCapture() as output:
            handler.output(result)

        self.assertIn('Pečené kuře', output.captured)
        result_data = {"msg": "Quagaars!", "code": 1000, "data": ['Pečené kuře']}
        self.assertEqual(json.loads(output.captured), result_data)
